@@ -1,6 +1,3 @@
-#import subprocess
-#subprocess.call(['pip', 'install', 'transformers'])
-
 import torch
 import numpy as np
 
@@ -9,7 +6,7 @@ from torch.utils.data import Dataset, DataLoader
 from pandas import DataFrame
 from typing import Optional, List
 
-from transformers import AutoTokenizer
+from transformers import AutoTokenizer, AutoModel
 
 #==================================================#
 #--------------------CONSTANTS---------------------#
@@ -123,7 +120,7 @@ class TokenizedDataset(Dataset):
       add_special_tokens=True,
       max_length=self.max_len,
       return_token_type_ids=False,
-      padding='max_length',
+      pad_to_max_length=True,
       return_attention_mask=True,
       return_tensors='pt'
     )
@@ -163,7 +160,8 @@ class EmbeddingModule(nn.Module):
   def __init__(self,
                basis : nn.Module,
                mode : Optional[int] = EmbModeConst.DEFAULT,
-               dropout : Optional[float] = 0):
+               dropout : Optional[float] = 0,
+               no_grad : Optional[bool] = False):
     
     super().__init__()
     self.add_module(ModuleConst.INP2EMB, basis)
@@ -172,14 +170,15 @@ class EmbeddingModule(nn.Module):
       self.add_module(ModuleConst.DROPOUT, nn.Dropout(p=dropout))
 
     self.mode = mode
+    self.no_grad = no_grad
     self.act_dict = {ModuleConst.INP2EMB : self.act_inp2emb,
                      ModuleConst.DROPOUT : self.act_dropout}
 
   def act_inp2emb(self, module, inp):
-    try:
-      emb = module(*inp)
-    except:
+    if isinstance(inp, torch.Tensor):
       emb = module(inp)
+    else:
+      emb = module(*inp)
 
     if self.mode in [EmbModeConst.DST_OUT, EmbModeConst.AGG_OUT]:
       if isinstance(emb, torch.Tensor):
@@ -197,7 +196,11 @@ class EmbeddingModule(nn.Module):
   def forward(self, inp):
     emb = inp
     for name, module in self.named_children():
-      emb = self.act_dict[name](module, emb)
+      if self.no_grad:
+        with torch.no_grad():
+          emb = self.act_dict[name](module, emb)
+      else:
+        emb = self.act_dict[name](module, emb)
     return emb
 
 #==================================================#
@@ -208,11 +211,13 @@ class ModuleChain():
   def __init__(self,
                chain : List[nn.Module],
                pre = lambda hid : hid,
-               post = lambda hid : hid):
+               post = lambda hid : hid,
+               no_grad : Optional[bool] = False):
     
     self._content = chain
     self._preprocess = pre
     self._postprocess = post
+    self._no_grad = no_grad
 
   @property
   def content(self):
@@ -226,15 +231,21 @@ class ModuleChain():
   def postprocess(self):
     return self._postprocess
 
+  @property
+  def no_grad(self):
+    return self._no_grad
+
 class ModuleChainFactory():  
   @staticmethod
   def get_linear_chain(dim_desc : List[int],
-                       bias_desc : Optional[List[bool]] = [True]
+                       bias_desc : Optional[List[bool]] = [True],
+                       no_grad : Optional[bool] = False
                        ) -> ModuleChain:
     chain = []
     for i in range(len(dim_desc) - 1):
       chain.append(nn.Linear(dim_desc[i], dim_desc[i + 1], bias_desc[i if i < len(bias_desc) else -1]))
-    return ModuleChain(chain)
+    return ModuleChain(chain, 
+                       no_grad = no_grad)
   
   RNN_PRE = {PreTransModeConst.RNN_STD : lambda hid : hid.unsqueeze(dim=1),
              PreTransModeConst.EMPTY   : lambda hid : hid}
@@ -250,7 +261,8 @@ class ModuleChainFactory():
                     pre_mode : Optional[int] = PreTransModeConst.RNN_STD,
                     post_mode : Optional[int] = PostTransModeConst.RNN_STD,
                     pre = None,
-                    post = None
+                    post = None,
+                    no_grad : Optional[bool] = False
                     ) -> ModuleChain:
     chain = []
     for i in range(len(dim_desc) - 1):
@@ -264,7 +276,8 @@ class ModuleChainFactory():
 
     return ModuleChain(chain, 
                        pre = pre if pre is not None else ModuleChainFactory.RNN_PRE[pre_mode], 
-                       post = post if post is not None else ModuleChainFactory.RNN_POST[post_mode])
+                       post = post if post is not None else ModuleChainFactory.RNN_POST[post_mode],
+                       no_grad = no_grad)
       
 class TransitionModule(nn.Module):
   def __init__(self,
@@ -306,7 +319,11 @@ class TransitionModule(nn.Module):
     for name, module in self.named_children():
       if name.startswith(ModuleConst.VEC2VEC):
         ids = name[len(ModuleConst.VEC2VEC):].split(ModuleConst.VECSPLIT)
-        vec = self.act_dict[ModuleConst.VEC2VEC](module, vec, int(ids[0]), int(ids[1]))
+        if self.chains[int(ids[0])].no_grad:
+          with torch.no_grad():
+            vec = self.act_dict[ModuleConst.VEC2VEC](module, vec, int(ids[0]), int(ids[1]))
+        else:
+          vec = self.act_dict[ModuleConst.VEC2VEC](module, vec, int(ids[0]), int(ids[1]))
       else:
         vec = self.act_dict[name](module, vec)
 
@@ -321,7 +338,8 @@ class AggregationModule(nn.Module):
                basis : Optional[nn.Module] = None,
                dropout : Optional[float] = 0,
                pre = lambda emb : emb,
-               post = lambda hid : hid):
+               post = lambda hid : hid,
+               no_grad : Optional[bool] = False):
     
     super().__init__()
     if basis is not None:
@@ -336,6 +354,7 @@ class AggregationModule(nn.Module):
     self._perform = lambda emb : post(pre(emb))
     self.preprocess = pre
     self.postprocess = post
+    self.no_grad = no_grad
     self.act_dict = {ModuleConst.EMB2HID : self.act_emb2hid,
                      ModuleConst.DROPOUT : self.act_dropout}
   @property
@@ -348,11 +367,10 @@ class AggregationModule(nn.Module):
 
   def act_emb2hid(self, module, emb):
     emb = self.preprocess(emb)
-    try:
-      hid, h0 = emb
-      hid = module(hid, h0)
-    except:
+    if isinstance(emb, torch.Tensor):
       hid = module(emb)
+    else:
+      hid = module(*emb)
     hid = self.postprocess(hid)
     return hid
   
@@ -364,7 +382,11 @@ class AggregationModule(nn.Module):
     if self.mode == AggModeConst.INR:
       hid = self.perform(hid)
     for name, module in self.named_children():
-        hid = self.act_dict[name](module, hid)
+        if self.no_grad:
+          with torch.no_grad():
+            hid = self.act_dict[name](module, hid)
+        else:
+          hid = self.act_dict[name](module, hid)
     return hid
 
 class AggregationModuleFactory():
@@ -376,10 +398,10 @@ class AggregationModuleFactory():
       self.dim = dim
       self.div = dim if div == 0 else div
     def __call__(self, hid):
-      try:
-        hid0, hid1 = hid 
-      except:
+      if isinstance(hid, torch.Tensor):
         hid0, hid1 = None, hid
+      else:
+        hid0, hid1 = hid 
       hid1 = torch.reshape(hid1, (self.dim, hid1.size(0), hid1.size(1) // self.div))
       return (hid0, hid1) if hid0 is not None else hid1
 
@@ -406,7 +428,8 @@ class AggregationModuleFactory():
                      pre_mode : Optional[int] = PreAggModeConst.EMPTY,
                      post_mode : Optional[int] = PostAggModeConst.LST_HID,
                      pre = None,
-                     post = None
+                     post = None,
+                     no_grad : Optional[bool] = False
                      ) -> AggregationModule:
 
     # BEGIN CHECK PRE
@@ -431,7 +454,8 @@ class AggregationModuleFactory():
                                                      batch_first = True),
                              dropout = dropout,
                              pre = pre if pre is not None else AggregationModuleFactory.RNN_PRE[pre_mode],
-                             post = post if post is not None else AggregationModuleFactory.RNN_POST[post_mode])
+                             post = post if post is not None else AggregationModuleFactory.RNN_POST[post_mode],
+                             no_grad = no_grad)
 
 #==================================================#
 #---------------SEMANTIC AGGREGATOR----------------#
@@ -471,18 +495,18 @@ class SemanticAggregator(nn.Module):
     return module(data)
 
   def act_intra(self, module, emb):
-    try:
-      res, inf = emb
-    except:
+    if isinstance(emb, torch.Tensor):
       res, inf = emb, None
+    else:
+      res, inf = emb
     res = module(res)
     return (res, inf) if inf is not None else res
 
   def act_incon(self, module, emb):
-    try:
-      inf, res = emb
-    except:
+    if isinstance(emb, torch.Tensor):
       inf, res = None, emb
+    else:
+      inf, res = emb
     res = module(res)
     return (inf, res) if inf is not None else res
 
